import { JSON_CONTENT_TYPE } from './constants'

/**
 * A wrapper around fetch() which performs an API request. Returns a Promise which is resolved with
 * the decoded JSON body of the response (unless method is DELETE) or which is rejected in case of
 * an error.
 */
export const apiFetch = <ResponseBody=any>(
  input: string | Request, init: RequestInit = {}, type: string = JSON_CONTENT_TYPE
): Promise<ResponseBody> => (
  fetch(input, {
    credentials: 'omit',  // do not send any cookies or auth credentials with request
    ...init,
    headers: {
      ...init.headers,
      'Content-Type': type,
    },
  })
  .then(response => {
    if(!response || !response.ok) {
      const errorText = `API returned error response: ${response.statusText}`;
      const error = { error: new Error(errorText), response };

      // Reject the call passing the response parsed as JSON, if necessary.
      return responseBody(response, type)
        .then(body => Promise.reject({...error, body: body as ResponseBody}))
        .catch(() => Promise.reject(error))
    }

    // Parse response body as JSON (unless it was a delete).
    if (init.method === 'DELETE') { return null; }

    return responseBody(response, type)
  })
);

/**
 * Append to a URL's query string based on properties from the passed object.
 */
export const appendQuery = (endpoint: string, o: {[index: string]: any} = {}): string => {
  const url = new URL(endpoint);
  Object.keys(o).forEach(key => {
    if (o[key] !== undefined) {
      if (Array.isArray(o[key])) {
        o[key].forEach((item:any) => url.searchParams.append(key, item));
      } else {
        url.searchParams.append(key, o[key])
      }
    }
  });
  return url.href;
};

/**
 * Return a promise to either the json or text of a request based on content type
 */
const responseBody = (response: Response, type:string) => (
  type===JSON_CONTENT_TYPE ? response.json() : response.text()
);


export default apiFetch;
