/**
 * Utilities for loading the google API client dynamically.
 *
 * See: https://developers.google.com/identity/sign-in/web/reference
 */

declare var gapi: any;

// Return a promise which is resolved when the top-level "gapi" namespace is available.
export const loadGapi = () => new Promise(resolve => {
  // The idea here is that we dynamically create an "on load" callback in the global window object
  // and add a <script> tag to the document body to load the Google API client. The "on load"
  // callback resolves the promise with the global gapi object.

  // Generate a random function name for the on load callback using some quite disgusting hackery
  // with the Math.random() function.
  const onLoadFunctionName = `__gapi_load_${Math.random().toString(36).substr(2)}`;

  // Create a function on the global window object which will be the on-load callback. After load
  // the window object will have a "gapi" property. We have to tell TypeScript to treat the window
  // object as a generic key/value object here so that it tolerates our hackery.
  (window as {[key: string]: any})[onLoadFunctionName] = resolve;

  // Add a <script> tag to load the Google API library
  const element = document.createElement('script') as HTMLScriptElement;
  element.src = `https://apis.google.com/js/platform.js?onload=${onLoadFunctionName}`;
  element.async = true;
  element.defer = true;
  document.body.appendChild(element);
})

// Return a Promise which is resolved when the top-level "gapi.auth2.init" function is available.
export const loadGapiAuth2 = async () => {
  await loadGapi();
  return new Promise((resolve, reject) => gapi.load('auth2', {
    callback: resolve,
    onerror: reject,
  }));
}
